"""Contains tests for the device handler classes."""

from __future__ import annotations

import asyncio
from datetime import timedelta
from math import isclose
from typing import Any, Literal, cast
from unittest.mock import AsyncMock, Mock, call, patch

import pytest

from pyplumio.const import (
    ATTR_DEVICE_INDEX,
    ATTR_FRAME_ERRORS,
    ATTR_INDEX,
    ATTR_OFFSET,
    ATTR_PARAMETER,
    ATTR_SCHEDULE,
    ATTR_SENSORS,
    ATTR_SETUP,
    ATTR_SIZE,
    ATTR_STATE,
    ATTR_SWITCH,
    ATTR_TYPE,
    ATTR_VALUE,
    STATE_OFF,
    STATE_ON,
    DeviceState,
    DeviceType,
    FrameType,
    UnitOfMeasurement,
)
from pyplumio.devices import get_device_handler
from pyplumio.devices.ecomax import (
    ATTR_FUEL_BURNED,
    ATTR_MIXERS,
    ATTR_THERMOSTATS,
    REQUIRED,
    EcoMAX,
)
from pyplumio.devices.ecoster import EcoSTER
from pyplumio.devices.mixer import Mixer
from pyplumio.devices.thermostat import Thermostat
from pyplumio.exceptions import RequestError, UnknownDeviceError
from pyplumio.frames import Response
from pyplumio.frames.messages import RegulatorDataMessage, SensorDataMessage
from pyplumio.frames.requests import (
    AlertsRequest,
    EcomaxControlRequest,
    RegulatorDataSchemaRequest,
    SchedulesRequest,
    SetEcomaxParameterRequest,
    SetMixerParameterRequest,
    SetScheduleRequest,
    SetThermostatParameterRequest,
    UIDRequest,
)
from pyplumio.frames.responses import (
    EcomaxParametersResponse,
    MixerParametersResponse,
    RegulatorDataSchemaResponse,
    SchedulesResponse,
    ThermostatParametersResponse,
)
from pyplumio.helpers.schedule import Schedule, ScheduleDay
from pyplumio.parameters.ecomax import (
    PARAMETER_TYPES as ECOMAX_PARAMETER_TYPES,
    EcomaxNumber,
    EcomaxSwitch,
)
from pyplumio.parameters.mixer import MixerNumber, MixerSwitch
from pyplumio.parameters.thermostat import ThermostatNumber
from pyplumio.structures.ecomax_parameters import ATTR_ECOMAX_CONTROL
from pyplumio.structures.frame_versions import ATTR_FRAME_VERSIONS
from pyplumio.structures.fuel_consumption import ATTR_FUEL_CONSUMPTION
from pyplumio.structures.mixer_parameters import ATTR_MIXER_PARAMETERS
from pyplumio.structures.mixer_sensors import ATTR_MIXER_SENSORS
from pyplumio.structures.network_info import NetworkInfo
from pyplumio.structures.regulator_data import ATTR_REGDATA
from pyplumio.structures.schedules import (
    ATTR_SCHEDULE_PARAMETER,
    ATTR_SCHEDULE_SWITCH,
    ATTR_SCHEDULES,
    ScheduleNumber,
    ScheduleSwitch,
)
from pyplumio.structures.thermostat_parameters import (
    ATTR_THERMOSTAT_PARAMETERS,
    ATTR_THERMOSTAT_PROFILE,
)
from pyplumio.structures.thermostat_sensors import (
    ATTR_THERMOSTAT_SENSORS,
    ATTR_THERMOSTATS_AVAILABLE,
    ATTR_THERMOSTATS_CONNECTED,
)
from tests.conftest import (
    DEFAULT_TOLERANCE,
    RAISES,
    UNDEFINED,
    class_from_json,
    json_test_data,
)


@pytest.mark.parametrize(
    ("device_type", "handler"),
    [
        (DeviceType.ECOMAX, "devices.ecomax.EcoMAX"),
        (DeviceType.ECONET, "devices.econet.EcoNET"),
        (DeviceType.ECOSTER, "devices.ecoster.EcoSTER"),
        (99, RAISES),
    ],
)
def test_device_handler(
    device_type: DeviceType | int, handler: str | Literal["raises"]
) -> None:
    """Test getting device handler class by device address."""
    if handler == RAISES:
        with pytest.raises(UnknownDeviceError):
            get_device_handler(device_type)
    else:
        assert get_device_handler(device_type) == handler


def test_ecoster() -> None:
    """Test ecoster instance."""
    ecoster = EcoSTER(asyncio.Queue(), network=NetworkInfo())
    assert isinstance(ecoster, EcoSTER)


@patch("pyplumio.devices.ecomax.EcoMAX.wait_for")
async def test_setup(mock_wait_for: AsyncMock, ecomax: EcoMAX) -> None:
    """Test requesting initial data frames."""
    with patch("pyplumio.devices.ecomax.EcoMAX.request") as mock_request:
        await ecomax.dispatch(ATTR_SETUP, True)
        await ecomax.wait_until_done()

    assert ecomax.get_nowait(ATTR_SETUP)
    assert ATTR_FRAME_ERRORS not in ecomax.data
    assert mock_request.await_count == len(REQUIRED)
    mock_wait_for.assert_awaited_once_with(ATTR_SENSORS)


@patch("pyplumio.devices.ecomax.EcoMAX.wait_for")
async def test_setup_error(mock_wait_for: AsyncMock, ecomax: EcoMAX) -> None:
    """Test with error during requesting initial data frames."""
    with patch(
        "pyplumio.devices.ecomax.EcoMAX.request",
        side_effect=RequestError("test", FrameType.REQUEST_ALERTS),
    ) as mock_request:
        await ecomax.dispatch(ATTR_SETUP, True)
        await ecomax.wait_until_done()

    assert ecomax.get_nowait(ATTR_SETUP)
    assert ecomax.data[ATTR_FRAME_ERRORS][0] == FrameType.REQUEST_ALERTS
    assert mock_request.await_count == len(REQUIRED)
    mock_wait_for.assert_awaited_once_with(ATTR_SENSORS)


@patch("asyncio.Queue.put_nowait")
@class_from_json(SensorDataMessage, "messages/sensor_data.json", arguments=("message",))
async def test_frame_versions_update(
    mock_put_nowait, ecomax: EcoMAX, sensor_data
) -> None:
    """Test requesting updated frames."""
    with patch("pyplumio.devices.ecomax.REQUIRED_TYPES", []):
        ecomax.handle_frame(sensor_data)
        await ecomax.wait_until_done()

    mock_put_nowait.assert_has_calls(
        [
            call(RegulatorDataSchemaRequest(recipient=DeviceType.ECOMAX)),
            call(SchedulesRequest(recipient=DeviceType.ECOMAX)),
            call(UIDRequest(recipient=DeviceType.ECOMAX)),
            call(AlertsRequest(recipient=DeviceType.ECOMAX)),
        ]
    )

    def _make_sensor_message_data(version: int) -> dict[str, Any]:
        """Create truncated sensor data containing only one version."""
        return {
            ATTR_SENSORS: {
                ATTR_FRAME_VERSIONS: {FrameType.REQUEST_REGULATOR_DATA_SCHEMA: version}
            }
        }

    # Test with frame type that are handled during setup.
    mock_put_nowait.reset_mock()
    sensor_data_message = SensorDataMessage(data=_make_sensor_message_data(86))
    ecomax.data.clear()
    ecomax.handle_frame(sensor_data_message)
    await ecomax.wait_until_done()
    mock_put_nowait.assert_not_called()

    # Test with same frame type after setup done.
    mock_put_nowait.reset_mock()
    sensor_data_message2 = SensorDataMessage(data=_make_sensor_message_data(87))
    ecomax.data.clear()
    ecomax.data[ATTR_SETUP] = True
    ecomax.handle_frame(sensor_data_message2)
    await ecomax.wait_until_done()
    mock_put_nowait.assert_called_once_with(
        RegulatorDataSchemaRequest(recipient=DeviceType.ECOMAX)
    )


@class_from_json(SensorDataMessage, "messages/sensor_data.json", arguments=("message",))
async def test_ecomax_data_callbacks(
    ecomax: EcoMAX, sensor_data: SensorDataMessage
) -> None:
    """Test callbacks dispatched on data frames."""
    ecomax.handle_frame(sensor_data)
    await ecomax.wait_until_done()
    heating_target = cast(float, ecomax.get_nowait("heating_target"))
    assert isclose(heating_target, 41.0, rel_tol=DEFAULT_TOLERANCE)

    ecomax_control = ecomax.get_nowait(ATTR_ECOMAX_CONTROL)
    assert isinstance(ecomax_control, EcomaxSwitch)
    assert ecomax_control.value == STATE_OFF

    ecomax_control_request = await ecomax_control.create_request()
    assert isinstance(ecomax_control_request, EcomaxControlRequest)
    assert ecomax_control_request.data == {ATTR_VALUE: 0}

    # Check ecomax_control reference equility.
    sensor_data_message2 = SensorDataMessage(data={ATTR_STATE: DeviceState.WORKING})
    ecomax.handle_frame(sensor_data_message2)
    await ecomax.wait_until_done()
    ecomax_control2 = ecomax.get_nowait(ATTR_ECOMAX_CONTROL)
    assert isinstance(ecomax_control2, EcomaxSwitch)
    assert ecomax_control2 is ecomax_control
    assert ecomax_control2.value == STATE_ON
    ecomax_control2_request = await ecomax_control.create_request()
    assert ecomax_control2_request.data == {ATTR_VALUE: 1}


@class_from_json(SensorDataMessage, "messages/sensor_data.json", arguments=("message",))
async def test_ecomax_naming_collisions(
    ecomax: EcoMAX, sensor_data: SensorDataMessage
) -> None:
    """Test that ecoMAX sensors don't collide with ecoMAX parameters."""
    ecomax.handle_frame(sensor_data)
    await ecomax.wait_until_done()
    for descriptions in ECOMAX_PARAMETER_TYPES.values():
        collisions = [
            description.name
            for description in descriptions
            if description.name in ecomax.data
        ]
        assert not collisions, f"found collisions: {collisions}"


@patch("asyncio.Queue.put")
@class_from_json(
    EcomaxParametersResponse, "responses/ecomax_parameters.json", arguments=("message",)
)
async def test_ecomax_parameters_callbacks(
    mock_put: AsyncMock,
    ecomax: EcoMAX,
    ecomax_parameters: EcomaxParametersResponse,
) -> None:
    """Test callbacks dispatched on parameter frames."""
    ecomax.handle_frame(ecomax_parameters)
    await ecomax.wait_until_done()
    fuzzy_logic = ecomax.get_nowait("fuzzy_logic")
    assert isinstance(fuzzy_logic, EcomaxSwitch)
    assert fuzzy_logic == STATE_ON
    assert fuzzy_logic.value == STATE_ON
    fuzzy_logic_request = await fuzzy_logic.create_request()
    assert isinstance(fuzzy_logic_request, SetEcomaxParameterRequest)
    assert fuzzy_logic_request.data == {
        ATTR_INDEX: 18,
        ATTR_VALUE: 1,
    }

    # Test that parameter instance is not recreated on subsequent calls.
    ecomax.handle_frame(ecomax_parameters)
    await ecomax.wait_until_done()
    assert ecomax.get_nowait("fuzzy_logic") is fuzzy_logic

    # Test parameter with the step (heating_heat_curve)
    fuel_calorific_value = ecomax.get_nowait("fuel_calorific_value")
    assert isinstance(fuel_calorific_value, EcomaxNumber)
    assert isclose(fuel_calorific_value.value, 4.6, rel_tol=DEFAULT_TOLERANCE)
    assert isclose(fuel_calorific_value.min_value, 0.1, rel_tol=DEFAULT_TOLERANCE)
    assert isclose(fuel_calorific_value.max_value, 25.0, rel_tol=DEFAULT_TOLERANCE)

    # Test setting parameter with the step.
    await fuel_calorific_value.set(2.5)
    request = mock_put.call_args[0][0]
    assert isinstance(request, SetEcomaxParameterRequest)
    assert request.data[ATTR_VALUE] == 25

    # Test parameter with the offset (heating_heat_curve_shift)
    heating_heat_curve_shift = ecomax.get_nowait("heating_curve_shift")
    assert isinstance(heating_heat_curve_shift, EcomaxNumber)
    assert isclose(heating_heat_curve_shift.value, 0.0, rel_tol=DEFAULT_TOLERANCE)
    assert isclose(heating_heat_curve_shift.min_value, -20.0, rel_tol=DEFAULT_TOLERANCE)
    assert isclose(heating_heat_curve_shift.max_value, 20.0, rel_tol=DEFAULT_TOLERANCE)
    assert heating_heat_curve_shift.unit_of_measurement == UnitOfMeasurement.CELSIUS

    # Test setting the parameter with the offset.
    await heating_heat_curve_shift.set(1)
    request = mock_put.call_args[0][0]
    assert isinstance(request, SetEcomaxParameterRequest)
    assert request.data[ATTR_VALUE] == 21


@class_from_json(
    EcomaxParametersResponse,
    "unknown/unknown_ecomax_parameter.json",
    arguments=("message",),
)
async def test_unknown_ecomax_parameter(
    ecomax: EcoMAX, caplog, unknown_ecomax_parameter: EcomaxParametersResponse
) -> None:
    """Test unknown ecoMAX parameter."""
    ecomax.handle_frame(unknown_ecomax_parameter)
    await ecomax.wait_until_done()
    assert "unknown ecoMAX parameter (139)" in caplog.text
    assert "ParameterValues(value=1, min_value=1, max_value=1)" in caplog.text
    assert "ecoMAX 350P2-ZF" in caplog.text


async def test_fuel_consumption_callbacks(frozen_time, caplog, ecomax: EcoMAX) -> None:
    """Test callbacks dispatched on fuel consumption."""
    frozen_time.tick(timedelta(seconds=10))
    ecomax.handle_frame(Response(data={ATTR_FUEL_CONSUMPTION: 3.6}))
    await ecomax.wait_until_done()
    fuel_burned = cast(float, ecomax.get_nowait(ATTR_FUEL_BURNED))
    assert isclose(fuel_burned, 0.01, rel_tol=DEFAULT_TOLERANCE)

    frozen_time.tick(timedelta(minutes=5, seconds=10))
    ecomax.handle_frame(Response(data={ATTR_FUEL_CONSUMPTION: 1}))
    await ecomax.wait_until_done()
    fuel_burned = cast(float, ecomax.get_nowait(ATTR_FUEL_BURNED))
    assert isclose(fuel_burned, 0.01, rel_tol=DEFAULT_TOLERANCE)
    assert "Skipping outdated fuel consumption" in caplog.text
    caplog.clear()

    frozen_time.tick(timedelta(seconds=10))
    ecomax.handle_frame(Response(data={ATTR_FUEL_CONSUMPTION: 7.2}))
    await ecomax.wait_until_done()
    assert "Skipping outdated fuel consumption" not in caplog.text
    fuel_burned = cast(float, ecomax.get_nowait(ATTR_FUEL_BURNED))
    assert isclose(fuel_burned, 0.02, rel_tol=DEFAULT_TOLERANCE)


@class_from_json(
    RegulatorDataSchemaResponse,
    "responses/regulator_data_schema.json",
    arguments=("message",),
)
@class_from_json(
    RegulatorDataMessage, "messages/regulator_data.json", arguments=("message",)
)
@json_test_data("messages/regulator_data.json", selector="data")
async def test_regdata_callbacks(
    ecomax: EcoMAX,
    regulator_data_schema: RegulatorDataSchemaResponse,
    regulator_data: RegulatorDataMessage,
    regulator_data_data,
) -> None:
    """Test that callbacks are dispatched on regdata."""
    ecomax.handle_frame(regulator_data_schema)
    await ecomax.wait_until_done()
    ecomax.handle_frame(regulator_data)
    await ecomax.wait_until_done()
    frame_versions = ATTR_FRAME_VERSIONS
    assert ecomax.get_nowait(ATTR_REGDATA) == regulator_data_data[ATTR_REGDATA]
    assert ecomax.get_nowait(frame_versions) == regulator_data_data[frame_versions]


@class_from_json(
    RegulatorDataMessage, "messages/regulator_data.json", arguments=("message",)
)
async def test_regdata_callbacks_without_schema(
    ecomax: EcoMAX, regulator_data: RegulatorDataMessage
) -> None:
    """Test callbacks dispatched on regdata without regdata schema."""
    ecomax.handle_frame(regulator_data)
    await ecomax.wait_until_done()
    assert ATTR_FRAME_VERSIONS in ecomax.data
    assert ATTR_REGDATA not in ecomax.data


@class_from_json(SensorDataMessage, "messages/sensor_data.json", arguments=("message",))
async def test_mixer_sensors_callbacks(
    ecomax: EcoMAX, sensor_data: SensorDataMessage
) -> None:
    """Test callbacks dispatched on mixer sensors info."""
    ecomax.handle_frame(sensor_data)
    await ecomax.wait_until_done()
    mixers = cast(dict[int, Mixer], ecomax.get_nowait(ATTR_MIXERS))
    assert isinstance(mixers, dict)
    assert len(mixers) == 1
    mixer = mixers[4]
    assert isinstance(mixer, Mixer)
    assert mixer.index == 4
    assert mixer.data == {
        "current_temp": 20.0,
        "target_temp": 40,
        "pump": False,
        "mixer_sensors": True,
    }


async def test_mixer_sensors_callbacks_without_mixers(ecomax: EcoMAX) -> None:
    """Test callbacks dispatched on mixer sensors without any mixers."""
    ecomax.handle_frame(MixerParametersResponse(data={ATTR_MIXER_SENSORS: {}}))
    await ecomax.wait_until_done()
    assert not ecomax.get_nowait(ATTR_MIXER_SENSORS, UNDEFINED)


@class_from_json(SensorDataMessage, "messages/sensor_data.json", arguments=("message",))
async def test_thermostat_sensors_callbacks(
    ecomax: EcoMAX, sensor_data: SensorDataMessage
) -> None:
    """Test callbacks dispatched on thermostat sensors info."""
    ecomax.handle_frame(sensor_data)
    await ecomax.wait_until_done()
    thermostats = cast(dict[int, Thermostat], ecomax.get_nowait(ATTR_THERMOSTATS))
    assert isinstance(thermostats, dict)
    assert len(thermostats) == 1
    thermostat = thermostats[0]
    assert isinstance(thermostat, Thermostat)
    assert thermostat.index == 0
    assert thermostat.data == {
        "state": 3,
        "current_temp": 43.5,
        "target_temp": 50.0,
        "contacts": True,
        "schedule": False,
        "thermostat_sensors": True,
    }
    assert ecomax.get_nowait(ATTR_THERMOSTATS_AVAILABLE) == 2
    assert ecomax.get_nowait(ATTR_THERMOSTATS_CONNECTED) == 1


async def test_thermostat_sensors_callbacks_without_thermostats(ecomax: EcoMAX) -> None:
    """Test callbacks dispatched on thermostats sensors without any thermostats."""
    ecomax.handle_frame(
        ThermostatParametersResponse(data={ATTR_THERMOSTAT_SENSORS: {}})
    )
    await ecomax.wait_until_done()
    assert not ecomax.get_nowait(ATTR_THERMOSTAT_SENSORS, UNDEFINED)


@class_from_json(
    ThermostatParametersResponse,
    "responses/thermostat_parameters.json",
    arguments=("message",),
)
async def test_thermostat_parameters_callbacks(
    ecomax: EcoMAX, thermostat_parameters: ThermostatParametersResponse
) -> None:
    """Test callbacks dispatched on thermostat parameters."""
    ecomax.handle_frame(Response(data={ATTR_THERMOSTATS_AVAILABLE: 3}))
    await ecomax.wait_until_done()
    ecomax.handle_frame(thermostat_parameters)
    await ecomax.wait_until_done()
    thermostats = cast(dict[int, Thermostat], ecomax.get_nowait(ATTR_THERMOSTATS))
    assert isinstance(thermostats, dict)
    assert len(thermostats) == 1
    thermostat = thermostats[0]
    assert isinstance(thermostat, Thermostat)
    assert len(thermostat.data) == 13
    party_target_temp = thermostat.get_nowait("party_target_temp")
    assert isinstance(party_target_temp, ThermostatNumber)
    assert isinstance(party_target_temp.device, Thermostat)
    assert isclose(party_target_temp.value, 22.0, rel_tol=DEFAULT_TOLERANCE)
    assert isclose(party_target_temp.min_value, 10.0, rel_tol=DEFAULT_TOLERANCE)
    assert isclose(party_target_temp.max_value, 35.0, rel_tol=DEFAULT_TOLERANCE)
    assert party_target_temp.offset == 0
    party_target_temp_request = await party_target_temp.create_request()
    assert isinstance(party_target_temp_request, SetThermostatParameterRequest)
    assert party_target_temp_request.data == {
        ATTR_INDEX: 2,
        ATTR_VALUE: 220,
        ATTR_OFFSET: 0,
        ATTR_SIZE: 2,
    }

    # Test that parameter instance is not recreated on subsequent calls.
    ecomax.handle_frame(thermostat_parameters)
    await ecomax.wait_until_done()
    thermostats = cast(dict[int, Thermostat], ecomax.get_nowait(ATTR_THERMOSTATS))
    assert isinstance(thermostats, dict)
    thermostat = thermostats[0]
    assert isinstance(thermostat, Thermostat)
    assert thermostat.get_nowait("party_target_temp") is party_target_temp


@class_from_json(
    ThermostatParametersResponse,
    "responses/thermostat_parameters.json",
    arguments=("message",),
)
async def test_thermostat_parameters_callbacks_without_thermostats(
    ecomax: EcoMAX, thermostat_parameters: ThermostatParametersResponse
) -> None:
    """Test callbacks dispatched on thermostat parameters without any thermostats."""
    ecomax.handle_frame(Response(data={ATTR_THERMOSTATS_AVAILABLE: 0}))
    ecomax.handle_frame(thermostat_parameters)
    await ecomax.wait_until_done()
    assert not ecomax.get_nowait(ATTR_THERMOSTAT_PARAMETERS, UNDEFINED)


@class_from_json(
    ThermostatParametersResponse,
    "responses/thermostat_parameters.json",
    arguments=("message",),
)
async def test_thermostat_profile_callbacks(
    ecomax: EcoMAX, thermostat_parameters: ThermostatParametersResponse
) -> None:
    """Test callbacks dispatched on thermostat profile."""
    ecomax.handle_frame(Response(data={ATTR_THERMOSTATS_AVAILABLE: 3}))
    await ecomax.wait_until_done()
    ecomax.handle_frame(thermostat_parameters)
    await ecomax.wait_until_done()
    thermostat_profile = ecomax.get_nowait(ATTR_THERMOSTAT_PROFILE)
    assert isinstance(thermostat_profile, EcomaxNumber)
    assert isclose(thermostat_profile.value, 0.0, rel_tol=DEFAULT_TOLERANCE)
    assert isclose(thermostat_profile.min_value, 0.0, rel_tol=DEFAULT_TOLERANCE)
    assert isclose(thermostat_profile.max_value, 5.0, rel_tol=DEFAULT_TOLERANCE)
    thermostat_profile_request = await thermostat_profile.create_request()
    assert isinstance(thermostat_profile_request, SetThermostatParameterRequest)
    assert thermostat_profile_request.data == {
        ATTR_INDEX: 0,
        ATTR_VALUE: 0,
        ATTR_OFFSET: 0,
        ATTR_SIZE: 1,
    }

    # Test when thermostat profile is none.
    ecomax.handle_frame(Response(data={ATTR_THERMOSTAT_PROFILE: None}))
    await ecomax.wait_until_done()
    assert ecomax.get_nowait(ATTR_THERMOSTAT_PROFILE) is None


@class_from_json(
    MixerParametersResponse,
    "responses/mixer_parameters.json",
    arguments=("message",),
)
async def test_mixer_parameters_callbacks(
    ecomax: EcoMAX, mixer_parameters: MixerParametersResponse
) -> None:
    """Test callbacks dispatched on mixer parameters."""
    ecomax.handle_frame(mixer_parameters)
    await ecomax.wait_until_done()
    mixers = cast(dict[int, Mixer], ecomax.get_nowait(ATTR_MIXERS))
    assert isinstance(mixers, dict)
    assert len(mixers) == 1
    mixer = mixers[0]
    assert isinstance(mixer, Mixer)
    assert len(mixer.data) == 7
    mixer_target_temp = mixer.get_nowait("mixer_target_temp")
    assert isinstance(mixer_target_temp, MixerNumber)
    assert isinstance(mixer_target_temp.device, Mixer)
    assert isclose(mixer_target_temp.value, 40.0, rel_tol=DEFAULT_TOLERANCE)
    assert isclose(mixer_target_temp.min_value, 30.0, rel_tol=DEFAULT_TOLERANCE)
    assert isclose(mixer_target_temp.max_value, 60.0, rel_tol=DEFAULT_TOLERANCE)

    # Test creating a request.
    mixer_target_temp_request = await mixer_target_temp.create_request()
    assert isinstance(mixer_target_temp_request, SetMixerParameterRequest)
    assert mixer_target_temp_request.data == {
        ATTR_INDEX: 0,
        ATTR_VALUE: 40,
        ATTR_DEVICE_INDEX: 0,
    }

    weather_control = mixer.get_nowait("weather_control")
    assert isinstance(weather_control, MixerSwitch)
    assert weather_control.value == STATE_ON

    heat_curve = mixer.get_nowait("heating_curve")
    assert isinstance(heat_curve, MixerNumber)
    assert isclose(heat_curve.value, 1.3, rel_tol=DEFAULT_TOLERANCE)
    assert isclose(heat_curve.min_value, 1.0, rel_tol=DEFAULT_TOLERANCE)
    assert isclose(heat_curve.max_value, 3.0, rel_tol=DEFAULT_TOLERANCE)

    # Test that parameter instance is not recreated on subsequent calls.
    ecomax.handle_frame(mixer_parameters)
    await ecomax.wait_until_done()
    mixers = cast(dict[int, Mixer], ecomax.get_nowait(ATTR_MIXERS))
    assert isinstance(mixers, dict)
    mixer = mixers[0]
    assert isinstance(mixer, Mixer)
    assert mixer.get_nowait("mixer_target_temp") is mixer_target_temp


async def test_mixer_parameters_callbacks_without_mixers(ecomax: EcoMAX) -> None:
    """Test mixer parameters callbacks without any mixers."""
    ecomax.handle_frame(MixerParametersResponse(data={ATTR_MIXER_PARAMETERS: {}}))
    await ecomax.wait_until_done()
    assert not ecomax.get_nowait(ATTR_MIXER_PARAMETERS, UNDEFINED)


@class_from_json(
    MixerParametersResponse,
    "unknown/unknown_mixer_parameter.json",
    arguments=("message",),
)
async def test_unknown_mixer_parameter(
    ecomax: EcoMAX, caplog, unknown_mixer_parameter: MixerParametersResponse
) -> None:
    """Test unknown mixer parameter."""
    ecomax.handle_frame(unknown_mixer_parameter)
    await ecomax.wait_until_done()
    assert "unknown mixer parameter (14)" in caplog.text
    assert "ParameterValues(value=1, min_value=1, max_value=1)" in caplog.text
    assert "ecoMAX 350P2-ZF" in caplog.text


@class_from_json(SchedulesResponse, "responses/schedules.json", arguments=("message",))
@json_test_data("responses/schedules.json", selector="data")
async def test_schedule_callback(
    ecomax: EcoMAX, schedules: SchedulesResponse, schedules_data
) -> None:
    """Test callbacks dispatched on schedule data."""
    ecomax.handle_frame(schedules)
    await ecomax.wait_until_done()
    schedules_arg = cast(dict[str, Schedule], ecomax.get_nowait(ATTR_SCHEDULES))
    assert isinstance(schedules_arg, dict)
    assert len(schedules_arg) == 2
    heating_schedule = schedules_arg["heating"]
    assert isinstance(heating_schedule, Schedule)

    heating_schedule_switch = ecomax.get_nowait("heating_schedule_switch")
    assert isinstance(heating_schedule_switch, ScheduleSwitch)
    assert heating_schedule_switch.value == STATE_OFF

    water_heater_schedule_parameter = ecomax.get_nowait(
        "water_heater_schedule_parameter"
    )
    assert isinstance(water_heater_schedule_parameter, ScheduleNumber)
    assert water_heater_schedule_parameter.value == 5
    assert water_heater_schedule_parameter.min_value == 0
    assert water_heater_schedule_parameter.max_value == 30
    water_heater_schedule_parameter_request = (
        await water_heater_schedule_parameter.create_request()
    )
    assert isinstance(water_heater_schedule_parameter_request, SetScheduleRequest)
    assert water_heater_schedule_parameter_request.data == {
        ATTR_TYPE: "water_heater",
        ATTR_SWITCH: ecomax.data[f"water_heater_{ATTR_SCHEDULE_SWITCH}"],
        ATTR_PARAMETER: ecomax.data[f"water_heater_{ATTR_SCHEDULE_PARAMETER}"],
        ATTR_SCHEDULE: ecomax.data[ATTR_SCHEDULES]["water_heater"],
    }

    schedule_data = schedules_data[ATTR_SCHEDULES][0][1]
    for index, weekday in enumerate(
        ("monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "sunday")
    ):
        schedule = getattr(heating_schedule, weekday)
        assert (
            schedule.schedule
            == ScheduleDay.from_iterable(schedule_data[index]).schedule
        )

    # Test that parameter instance is not recreated on subsequent calls.
    ecomax.handle_frame(schedules)
    await ecomax.wait_until_done()
    assert ecomax.get_nowait("heating_schedule_switch") is heating_schedule_switch


async def test_request(ecomax: EcoMAX) -> None:
    """Test requesting the value."""
    with (
        patch(
            "pyplumio.devices.ecomax.EcoMAX.get",
            side_effect=(asyncio.TimeoutError, True),
        ) as mock_get,
        patch("asyncio.Queue.put_nowait") as mock_put_nowait,
    ):
        assert await ecomax.request("foo", FrameType.REQUEST_ALERTS, timeout=5)

    mock_get.assert_awaited_with("foo", timeout=5)
    assert mock_put_nowait.call_count == 2
    args, _ = mock_put_nowait.call_args
    assert isinstance(args[0], AlertsRequest)


async def test_request_error(ecomax: EcoMAX) -> None:
    """Test requesting the value with error."""
    with (
        patch(
            "pyplumio.devices.ecomax.EcoMAX.get",
            side_effect=(asyncio.TimeoutError, asyncio.TimeoutError),
        ),
        pytest.raises(RequestError) as excinfo,
    ):
        await ecomax.request("foo", FrameType.REQUEST_ALERTS, retries=1)

    assert (
        excinfo.value.args[0] == "Failed to request 'foo' with frame type '61' "
        "after 0 retries."
    )
    assert excinfo.value.frame_type == FrameType.REQUEST_ALERTS


@json_test_data("responses/ecomax_parameters.json", selector="message")
@json_test_data("responses/thermostat_parameters.json", selector="message")
@json_test_data("responses/mixer_parameters.json", selector="message")
@patch("pyplumio.parameters.Parameter.pending_update", False)
async def test_set(
    ecomax: EcoMAX,
    ecomax_parameters_message,
    thermostat_parameters_message,
    mixer_parameters_message,
) -> None:
    """Test setting parameter value via set helper."""
    ecomax.handle_frame(Response(data={ATTR_THERMOSTATS_AVAILABLE: 3}))
    await ecomax.wait_until_done()
    ecomax.handle_frame(EcomaxParametersResponse(message=ecomax_parameters_message))
    ecomax.handle_frame(
        ThermostatParametersResponse(message=thermostat_parameters_message)
    )
    ecomax.handle_frame(MixerParametersResponse(message=mixer_parameters_message))
    await ecomax.wait_until_done()

    # Test setting an ecomax parameter.
    assert await ecomax.set("max_fuel_flow", 26.0)
    max_fuel_flow = ecomax.get_nowait("max_fuel_flow")
    assert isinstance(max_fuel_flow, EcomaxNumber)
    assert isclose(max_fuel_flow.value, 26.0, rel_tol=DEFAULT_TOLERANCE)

    # Test setting an ecomax parameter with invalid step.
    with pytest.raises(ValueError):
        await ecomax.set("max_fuel_flow", 26.1)

    # Test setting an ecomax parameter without blocking.
    with (
        patch("pyplumio.devices.Device.set", new_callable=Mock),
        patch(
            "pyplumio.helpers.task_manager.TaskManager.create_task"
        ) as mock_create_task,
    ):
        ecomax.set_nowait("max_fuel_flow", 20)

    mock_create_task.assert_called_once()

    # Test setting a thermostat parameter.
    thermostat = ecomax.data[ATTR_THERMOSTATS][0]
    assert isinstance(thermostat, Thermostat)
    assert await thermostat.set("party_target_temp", 21)
    target_party_temp = thermostat.get_nowait("party_target_temp")
    assert isinstance(target_party_temp, ThermostatNumber)
    assert isclose(target_party_temp.value, 21.0, rel_tol=DEFAULT_TOLERANCE)

    # Test setting a thermostat parameter with invalid step.
    with pytest.raises(ValueError):
        await thermostat.set("party_target_temp", 26.01)

    # Test setting a mixer parameter.
    mixer = ecomax.data[ATTR_MIXERS][0]
    assert isinstance(mixer, Mixer)
    assert await mixer.set("mixer_target_temp", 35.0)
    mixer_target_temp = mixer.get_nowait("mixer_target_temp")
    assert isinstance(mixer_target_temp, MixerNumber)
    assert isclose(mixer_target_temp.value, 35.0, rel_tol=DEFAULT_TOLERANCE)

    # Test setting a mixer parameter with invalid step.
    with pytest.raises(ValueError):
        await mixer.set("mixer_target_temp", 35.01)

    # Test with invalid parameter.
    ecomax.data["bar"] = Mock()
    with pytest.raises(TypeError):
        await ecomax.set("bar", 1)


async def test_turn_on(ecomax: EcoMAX, caplog) -> None:
    """Test turning the controller on."""
    assert not await ecomax.turn_on()
    assert "ecoMAX control is not available. Please try again later." in caplog.text
    ecomax.data[ATTR_ECOMAX_CONTROL] = AsyncMock(spec=EcomaxSwitch)
    assert await ecomax.turn_on()
    ecomax.data[ATTR_ECOMAX_CONTROL].set.assert_awaited_once_with(STATE_ON)


async def test_turn_off(ecomax: EcoMAX, caplog) -> None:
    """Test turning the controller off."""
    await ecomax.turn_off()
    assert "ecoMAX control is not available. Please try again later." in caplog.text
    ecomax.data[ATTR_ECOMAX_CONTROL] = AsyncMock()
    await ecomax.turn_off()
    ecomax.data[ATTR_ECOMAX_CONTROL].set.assert_awaited_once_with(STATE_OFF)


@patch("pyplumio.devices.ecomax.EcoMAX.create_task")
@patch("pyplumio.devices.ecomax.EcoMAX.turn_on", new_callable=Mock)
async def test_turn_on_nowait(mock_create_task, mock_turn_on, ecomax: EcoMAX) -> None:
    """Test turning the controller on without waiting."""
    ecomax.turn_on_nowait()
    await ecomax.wait_until_done()
    mock_create_task.assert_called_once()
    mock_turn_on.assert_called_once()


@patch("pyplumio.devices.ecomax.EcoMAX.create_task")
@patch("pyplumio.devices.ecomax.EcoMAX.turn_off", new_callable=Mock)
async def test_turn_off_nowait(
    mock_create_task: Mock, mock_turn_off: Mock, ecomax: EcoMAX
) -> None:
    """Test turning the controller on without waiting."""
    ecomax.turn_off_nowait()
    await ecomax.wait_until_done()
    mock_create_task.assert_called_once()
    mock_turn_off.assert_called_once()


@json_test_data("messages/sensor_data.json", selector="message")
async def test_shutdown(ecomax: EcoMAX, sensor_data_message) -> None:
    """Test device tasks shutdown."""
    ecomax.handle_frame(SensorDataMessage(message=sensor_data_message))
    await ecomax.wait_until_done()

    with (
        patch("pyplumio.devices.mixer.Mixer.shutdown") as mock_mixer_shutdown,
        patch(
            "pyplumio.devices.thermostat.Thermostat.shutdown"
        ) as mock_thermostat_shutdown,
        patch("pyplumio.devices.Device.cancel_tasks") as mock_cancel_tasks,
        patch("pyplumio.devices.Device.wait_until_done") as mock_wait_until_done,
    ):
        await ecomax.shutdown()

    mock_wait_until_done.assert_awaited_once()
    mock_cancel_tasks.assert_called_once()
    mock_thermostat_shutdown.assert_awaited_once()
    mock_mixer_shutdown.assert_awaited_once()
